package cn.com.qmth.examcloud.web.actuator;

import java.util.Collections;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedMap;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;

import com.codahale.metrics.Clock;
import com.codahale.metrics.Counter;
import com.codahale.metrics.Gauge;
import com.codahale.metrics.Histogram;
import com.codahale.metrics.Meter;
import com.codahale.metrics.MetricAttribute;
import com.codahale.metrics.MetricFilter;
import com.codahale.metrics.MetricRegistry;
import com.codahale.metrics.ScheduledReporter;
import com.codahale.metrics.Snapshot;
import com.codahale.metrics.Timer;
import com.google.common.collect.Lists;

/**
 * 数据报告
 *
 * @author WANGWEI
 * @date 2019年7月25日
 * @Copyright (c) 2018-? http://qmth.com.cn All Rights Reserved.
 */
public class DataReportor extends ScheduledReporter {
	/**
	 * Returns a new {@link Builder} for {@link DataReportor}.
	 *
	 * @param registry
	 *            the registry to report
	 * @return a {@link Builder} instance for a {@link DataReportor}
	 */
	public static Builder forRegistry(MetricRegistry registry) {
		return new Builder(registry);
	}

	/**
	 * A builder for {@link DataReportor} instances. Defaults to using the
	 * default locale and time zone, writing to {@code System.out}, converting
	 * rates to events/second, converting durations to milliseconds, and not
	 * filtering metrics.
	 */
	public static class Builder {
		private final MetricRegistry registry;

		private Clock clock;

		private TimeUnit rateUnit;

		private TimeUnit durationUnit;

		private MetricFilter filter;

		private ScheduledExecutorService executor;

		private boolean shutdownExecutorOnStop;

		private Set<MetricAttribute> disabledMetricAttributes;

		private Builder(MetricRegistry registry) {
			this.registry = registry;
			this.clock = Clock.defaultClock();
			this.rateUnit = TimeUnit.SECONDS;
			this.durationUnit = TimeUnit.MILLISECONDS;
			this.filter = MetricFilter.ALL;
			this.executor = null;
			this.shutdownExecutorOnStop = true;
			disabledMetricAttributes = Collections.emptySet();
		}

		/**
		 * Specifies whether or not, the executor (used for reporting) will be
		 * stopped with same time with reporter. Default value is true. Setting
		 * this parameter to false, has the sense in combining with providing
		 * external managed executor via
		 * {@link #scheduleOn(ScheduledExecutorService)}.
		 *
		 * @param shutdownExecutorOnStop
		 *            if true, then executor will be stopped in same time with
		 *            this reporter
		 * @return {@code this}
		 */
		public Builder shutdownExecutorOnStop(boolean shutdownExecutorOnStop) {
			this.shutdownExecutorOnStop = shutdownExecutorOnStop;
			return this;
		}

		/**
		 * Specifies the executor to use while scheduling reporting of metrics.
		 * Default value is null. Null value leads to executor will be auto
		 * created on start.
		 *
		 * @param executor
		 *            the executor to use while scheduling reporting of metrics.
		 * @return {@code this}
		 */
		public Builder scheduleOn(ScheduledExecutorService executor) {
			this.executor = executor;
			return this;
		}

		/**
		 * Use the given {@link Clock} instance for the time.
		 *
		 * @param clock
		 *            a {@link Clock} instance
		 * @return {@code this}
		 */
		public Builder withClock(Clock clock) {
			this.clock = clock;
			return this;
		}

		/**
		 * Convert rates to the given time unit.
		 *
		 * @param rateUnit
		 *            a unit of time
		 * @return {@code this}
		 */
		public Builder convertRatesTo(TimeUnit rateUnit) {
			this.rateUnit = rateUnit;
			return this;
		}

		/**
		 * Convert durations to the given time unit.
		 *
		 * @param durationUnit
		 *            a unit of time
		 * @return {@code this}
		 */
		public Builder convertDurationsTo(TimeUnit durationUnit) {
			this.durationUnit = durationUnit;
			return this;
		}

		/**
		 * Only report metrics which match the given filter.
		 *
		 * @param filter
		 *            a {@link MetricFilter}
		 * @return {@code this}
		 */
		public Builder filter(MetricFilter filter) {
			this.filter = filter;
			return this;
		}

		/**
		 * Don't report the passed metric attributes for all metrics (e.g.
		 * "p999", "stddev" or "m15"). See {@link MetricAttribute}.
		 *
		 * @param disabledMetricAttributes
		 *            a {@link MetricFilter}
		 * @return {@code this}
		 */
		public Builder disabledMetricAttributes(Set<MetricAttribute> disabledMetricAttributes) {
			this.disabledMetricAttributes = disabledMetricAttributes;
			return this;
		}

		/**
		 * Builds a {@link DataReportor} with the given properties.
		 *
		 * @return a {@link DataReportor}
		 */
		public DataReportor build() {
			return new DataReportor(registry, clock, rateUnit, durationUnit, filter, executor,
					shutdownExecutorOnStop, disabledMetricAttributes);
		}
	}

	private final Clock clock;

	private List<ReportInfo> reportInfoList = Lists.newArrayList();

	private int maxReportInfoSize = 10;

	private DataReportor(MetricRegistry registry, Clock clock, TimeUnit rateUnit,
			TimeUnit durationUnit, MetricFilter filter, ScheduledExecutorService executor,
			boolean shutdownExecutorOnStop, Set<MetricAttribute> disabledMetricAttributes) {
		super(registry, "data-reporter", filter, rateUnit, durationUnit, executor,
				shutdownExecutorOnStop, disabledMetricAttributes);
		this.clock = clock;
	}

	@Override
	@SuppressWarnings("rawtypes")
	public void report(SortedMap<String, Gauge> gauges, SortedMap<String, Counter> counters,
			SortedMap<String, Histogram> histograms, SortedMap<String, Meter> meters,
			SortedMap<String, Timer> timers) {

		ReportInfo info = new ReportInfo();
		info.setDateTime(new Date(clock.getTime()));
		final List<MeterInfo> meterInfoList = Lists.newArrayList();
		final List<TimerInfo> timerInfoList = Lists.newArrayList();
		info.setMeterInfoList(meterInfoList);
		info.setTimerInfoList(timerInfoList);

		if (!meters.isEmpty()) {
			for (Map.Entry<String, Meter> entry : meters.entrySet()) {
				MeterInfo meterInfo = getMeterInfo(entry.getKey(), entry.getValue());
				meterInfoList.add(meterInfo);
			}
		}

		if (!timers.isEmpty()) {
			for (Map.Entry<String, Timer> entry : timers.entrySet()) {
				TimerInfo timerInfo = getTimerInfo(entry.getKey(), entry.getValue());
				timerInfoList.add(timerInfo);
			}
		}

		if (reportInfoList.size() >= maxReportInfoSize) {
			reportInfoList = reportInfoList.subList(0, maxReportInfoSize);
		}

		reportInfoList.add(0, info);
	}

	private MeterInfo getMeterInfo(String key, Meter meter) {

		MeterInfo info = new MeterInfo();
		info.setKey(key);
		info.setRateUnit(getRateUnit());

		if (isEnabled(MetricAttribute.COUNT)) {
			info.setCount(meter.getCount());
		}
		if (isEnabled(MetricAttribute.MEAN_RATE)) {
			info.setMeanRate(meter.getMeanRate());
		}

		if (isEnabled(MetricAttribute.M1_RATE)) {
			info.setOneMinuteRate(convertRate(meter.getOneMinuteRate()));
		}
		if (isEnabled(MetricAttribute.M5_RATE)) {
			info.setFiveMinuteRate(convertRate(meter.getFiveMinuteRate()));
		}
		if (isEnabled(MetricAttribute.M15_RATE)) {
			info.setFifteenMinuteRate(convertRate(meter.getFifteenMinuteRate()));
		}

		return info;
	}

	private TimerInfo getTimerInfo(String key, Timer timer) {
		final Snapshot snapshot = timer.getSnapshot();
		TimerInfo info = new TimerInfo();
		info.setKey(key);
		info.setRateUnit(getRateUnit());
		info.setDurationUnit(getDurationUnit());

		if (isEnabled(MetricAttribute.COUNT)) {
			info.setCount(timer.getCount());
		}
		if (isEnabled(MetricAttribute.MEAN_RATE)) {
			info.setMeanRate(timer.getMeanRate());
		}

		if (isEnabled(MetricAttribute.M1_RATE)) {
			info.setOneMinuteRate(convertRate(timer.getOneMinuteRate()));
		}
		if (isEnabled(MetricAttribute.M5_RATE)) {
			info.setFiveMinuteRate(convertRate(timer.getFiveMinuteRate()));
		}
		if (isEnabled(MetricAttribute.M15_RATE)) {
			info.setFifteenMinuteRate(convertRate(timer.getFifteenMinuteRate()));
		}

		if (isEnabled(MetricAttribute.MIN)) {
			info.setMin(convertDuration(snapshot.getMin()));
		}
		if (isEnabled(MetricAttribute.MAX)) {
			info.setMax(convertDuration(snapshot.getMax()));
		}
		if (isEnabled(MetricAttribute.MEAN)) {
			info.setMean(convertDuration(snapshot.getMean()));
		}

		if (isEnabled(MetricAttribute.P50)) {
			info.setP50(convertDuration(snapshot.getMedian()));
		}
		if (isEnabled(MetricAttribute.P75)) {
			info.setP75(convertDuration(snapshot.get75thPercentile()));
		}
		if (isEnabled(MetricAttribute.P95)) {
			info.setP95(convertDuration(snapshot.get95thPercentile()));
		}
		if (isEnabled(MetricAttribute.P98)) {
			info.setP98(convertDuration(snapshot.get98thPercentile()));
		}
		if (isEnabled(MetricAttribute.P99)) {
			info.setP99(convertDuration(snapshot.get99thPercentile()));
		}
		if (isEnabled(MetricAttribute.P999)) {
			info.setP999(convertDuration(snapshot.get999thPercentile()));
		}

		return info;
	}

	private boolean isEnabled(MetricAttribute type) {
		return !getDisabledMetricAttributes().contains(type);
	}

	public List<ReportInfo> getReportInfoList() {
		return reportInfoList;
	}

}
